@@ -1,5 +1,5 @@
 /* Tarlz - Archiver with multimember lzip compression
-   Copyright (C) 2013-2023 Antonio Diaz Diaz.
+   Copyright (C) 2013-2025 Antonio Diaz Diaz.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -22,9 +22,7 @@
 #include <vector>
 #include <stdint.h>
 
-#define max_file_size ( LLONG_MAX - header_size )
-enum { header_size = 512,
-       max_edata_size = ( INT_MAX / header_size - 2 ) * header_size };
+enum { header_size = 512 };
 typedef uint8_t Tar_header[header_size];
 
 enum Offsets {
@@ -81,9 +79,9 @@ inline int decimal_digits( unsigned long long value )
 
 inline bool dotdot_at_i( const char * const filename, const int i )
   {
-  return ( filename[i] == '.' && filename[i+1] == '.' &&
-           ( i == 0 || filename[i-1] == '/' ) &&
-           ( filename[i+2] == 0 || filename[i+2] == '/' ) );
+  return filename[i] == '.' && filename[i+1] == '.' &&
+         ( i == 0 || filename[i-1] == '/' ) &&
+         ( filename[i+2] == 0 || filename[i+2] == '/' );
   }
 
 
@@ -120,7 +118,7 @@ public:
       }
     return true;
     }
-  char * operator()() { return p; }
+  char * operator()() { return p; }		// no need for operator[]
   const char * operator()() const { return p; }
   uint8_t * u8() { return (uint8_t *)p; }
   const uint8_t * u8() const { return (const uint8_t *)p; }
@@ -184,9 +182,12 @@ class Extended			// stores metadata from/for extended records
   mutable bool crc_present_;
 
   void calculate_sizes() const;
-  void unknown_keyword( const char * const buf, const int size ) const;
+  void unknown_keyword( const char * const buf, const int size,
+                        std::vector< std::string > * const msg_vecp = 0 ) const;
 
 public:
+  enum { max_edata_size = ( INT_MAX / header_size - 2 ) * header_size };
+  enum { max_file_size = LLONG_MAX - header_size };	// for padding
   static const std::string crc_record;
   std::string removed_prefix;
 
@@ -234,7 +235,8 @@ public:
 
   bool crc_present() const { return crc_present_; }
   bool parse( const char * const buf, const int edsize,
-              const bool permissive );
+              const bool permissive,
+              std::vector< std::string > * const msg_vecp = 0 );
   void fill_from_ustar( const Tar_header header );
   };
 
@@ -244,10 +246,10 @@ class CRC32
   uint32_t data[256];		// Table of CRCs of all 8-bit messages.
 
 public:
-  CRC32( const bool castagnoli = false )
+  explicit CRC32( const bool castagnoli = false )
     {
-    const unsigned cpol = 0x82F63B78U;	// CRC32-C  Castagnoli polynomial.
-    const unsigned ipol = 0xEDB88320U;	// IEEE 802.3 Ethernet polynomial.
+    const unsigned cpol = 0x82F63B78U;	// CRC32-C  Castagnoli polynomial
+    const unsigned ipol = 0xEDB88320U;	// IEEE 802.3 Ethernet polynomial
     const unsigned poly = castagnoli ? cpol : ipol;
 
     for( unsigned n = 0; n < 256; ++n )
@@ -275,12 +277,11 @@ public:
   uint32_t compute_crc( const uint8_t * const buffer, const int size ) const
     {
     uint32_t crc = 0xFFFFFFFFU;
-    for( int i = 0; i < size; ++i )
-      crc = data[(crc^buffer[i])&0xFF] ^ ( crc >> 8 );
+    update_buf( crc, buffer, size );
     return crc ^ 0xFFFFFFFFU;
     }
 
-  // Calculates the crc of size bytes except a window of 8 bytes at pos
+  // compute the crc of size bytes except a window of 8 bytes at pos
   uint32_t windowed_crc( const uint8_t * const buffer, const int pos,
                          const int size ) const
     {
@@ -322,8 +323,8 @@ enum {
 
 
 inline bool isvalid_ds( const unsigned dictionary_size )
-  { return ( dictionary_size >= min_dictionary_size &&
-             dictionary_size <= max_dictionary_size ); }
+  { return dictionary_size >= min_dictionary_size &&
+           dictionary_size <= max_dictionary_size; }
 
 
 const uint8_t lzip_magic[4] = { 0x4C, 0x5A, 0x49, 0x50 };	// "LZIP"
@@ -335,14 +336,13 @@ struct Lzip_header
 					//   4 version
 					//   5 coded dictionary size
 
-  bool check_magic() const
-    { return ( std::memcmp( data, lzip_magic, 4 ) == 0 ); }
+  bool check_magic() const { return std::memcmp( data, lzip_magic, 4 ) == 0; }
 
   bool check_prefix( const int sz ) const	// detect (truncated) header
     {
     for( int i = 0; i < sz && i < 4; ++i )
       if( data[i] != lzip_magic[i] ) return false;
-    return ( sz > 0 );
+    return sz > 0;
     }
 
   bool check_corrupt() const			// detect corrupt header
@@ -350,15 +350,15 @@ struct Lzip_header
     int matches = 0;
     for( int i = 0; i < 4; ++i )
       if( data[i] == lzip_magic[i] ) ++matches;
-    return ( matches > 1 && matches < 4 );
+    return matches > 1 && matches < 4;
     }
 
   uint8_t version() const { return data[4]; }
-  bool check_version() const { return ( data[4] == 1 ); }
+  bool check_version() const { return data[4] == 1; }
 
   unsigned dictionary_size() const
     {
-    unsigned sz = ( 1 << ( data[5] & 0x1F ) );
+    unsigned sz = 1 << ( data[5] & 0x1F );
     if( sz > min_dictionary_size )
       sz -= ( sz / 16 ) * ( ( data[5] >> 5 ) & 7 );
     return sz;
@@ -419,7 +419,7 @@ enum Program_mode { m_none, m_append, m_compress, m_concatenate, m_create,
 enum Solidity { no_solid, bsolid, dsolid, asolid, solid };
 class Arg_parser;
 
-struct Cl_options		// command line options
+struct Cl_options		// command-line options
   {
   const Arg_parser & parser;
   std::string archive_name;
@@ -435,26 +435,36 @@ struct Cl_options		// command line options
   int num_files;
   int num_workers;		// start this many worker threads
   int out_slots;
+  bool depth;
   bool dereference;
   bool filenames_given;
   bool ignore_ids;
+  bool ignore_metadata;
   bool ignore_overflow;
   bool keep_damaged;
   bool level_set;		// compression level set in command line
   bool missing_crc;
+  bool mount;
   bool mtime_set;
+  bool option_C_present;
+  bool option_T_present;
+  bool parallel;
   bool permissive;
   bool preserve_permissions;
+  bool recursive;
   bool warn_newer;
+  bool xdev;
 
   Cl_options( const Arg_parser & ap )
     : parser( ap ), mtime( 0 ), uid( -1 ), gid( -1 ), program_mode( m_none ),
       solidity( bsolid ), data_size( 0 ), debug_level( 0 ), level( 6 ),
-      num_files( 0 ), num_workers( -1 ), out_slots( 64 ), dereference( false ),
-      filenames_given( false ), ignore_ids( false ), ignore_overflow( false ),
-      keep_damaged( false ), level_set( false ), missing_crc( false ),
-      mtime_set( false ), permissive( false ), preserve_permissions( false ),
-      warn_newer( false ) {}
+      num_files( 0 ), num_workers( -1 ), out_slots( 64 ), depth( false ),
+      dereference( false ), filenames_given( false ), ignore_ids( false ),
+      ignore_metadata( false ), ignore_overflow( false ), keep_damaged( false ),
+      level_set( false ), missing_crc( false ), mount( false ),
+      mtime_set( false ), option_C_present( false ), option_T_present( false ),
+      parallel( false ), permissive( false ), preserve_permissions( false ),
+      recursive( true ), warn_newer( false ), xdev( false ) {}
 
   void set_level( const int l ) { level = l; level_set = true; }
 
@@ -469,13 +479,13 @@ inline void set_retval( int & retval, const int new_val )
 const char * const bad_magic_msg = "Bad magic number (file not in lzip format).";
 const char * const bad_dict_msg = "Invalid dictionary size in member header.";
 const char * const corrupt_mm_msg = "Corrupt header in multimember file.";
-const char * const trailing_msg = "Trailing data not allowed.";
 const char * const bad_hdr_msg = "Corrupt or invalid tar header.";
 const char * const gblrec_msg = "Error in global extended records.";
 const char * const extrec_msg = "Error in extended records.";
 const char * const miscrc_msg = "Missing CRC in extended records.";
 const char * const misrec_msg = "Missing extended records.";
 const char * const longrec_msg = "Extended records are too long.";
+const char * const large_file_msg = "Input file is too large.";
 const char * const end_msg = "Archive ends unexpectedly.";
 const char * const mem_msg = "Not enough memory.";
 const char * const mem_msg2 = "Not enough memory. Try a lower compression level.";
@@ -487,25 +497,21 @@ const char * const posix_lz_msg = "This does not look like a POSIX tar.lz archiv
 const char * const eclosa_msg = "Error closing archive";
 const char * const eclosf_msg = "Error closing file";
 const char * const nfound_msg = "Not found in archive.";
+const char * const rd_open_msg = "Can't open for reading";
+const char * const rd_err_msg = "Read error";
+const char * const wr_err_msg = "Write error";
 const char * const seek_msg = "Seek error";
-const char * const werr_msg = "Write error";
 const char * const chdir_msg = "Error changing working directory";
+const char * const intdir_msg = "Failed to create intermediate directory";
 
 // defined in common.cc
 unsigned long long parse_octal( const uint8_t * const ptr, const int size );
-int readblock( const int fd, uint8_t * const buf, const int size );
+long readblock( const int fd, uint8_t * const buf, const long size );
 int writeblock( const int fd, const uint8_t * const buf, const int size );
 
 // defined in common_decode.cc
 bool block_is_zero( const uint8_t * const buf, const int size );
-bool format_member_name( const Extended & extended, const Tar_header header,
-                         Resizable_buffer & rbuf, const bool long_format );
-bool show_member_name( const Extended & extended, const Tar_header header,
-                       const int vlevel, Resizable_buffer & rbuf );
-bool check_skip_filename( const Cl_options & cl_opts,
-                          std::vector< char > & name_pending,
-                          const char * const filename, const int chdir_fd = -1 );
-bool make_path( const std::string & name );
+bool make_dirs( const std::string & name );
 
 // defined in common_mutex.cc
 void exit_fail_mt( const int retval = 1 );	// terminate the program
@@ -520,14 +526,16 @@ void show_atpos_error( const char * const filename, const long long pos,
 int compress( const Cl_options & cl_opts );
 
 // defined in create.cc
-bool copy_file( const int infd, const int outfd, const long long max_size = -1 );
+bool copy_file( const int infd, const int outfd, const char * const filename,
+                const long long max_size = -1 );
 bool writeblock_wrapper( const int outfd, const uint8_t * const buffer,
                          const int size );
 bool write_eoa_records( const int outfd, const bool compressed );
 const char * remove_leading_dotslash( const char * const filename,
              std::string * const removed_prefixp, const bool dotdot = false );
-bool fill_headers( const char * const filename, Extended & extended,
-                   Tar_header header, long long & file_size, const int flag );
+bool fill_headers( std::string & estr, const char * const filename,
+                   Extended & extended, Tar_header header,
+                   long long & file_size, const int flag );
 bool block_is_full( const int extended_size,
                     const unsigned long long file_size,
                     const unsigned long long target_size,
@@ -538,10 +546,6 @@ bool has_lz_ext( const std::string & name );
 int concatenate( const Cl_options & cl_opts );
 int encode( const Cl_options & cl_opts );
 
-// defined in create_lz.cc
-int encode_lz( const Cl_options & cl_opts, const char * const archive_namep,
-               const int outfd );
-
 // defined in decode.cc
 bool compare_file_type( std::string & estr, std::string & ostr,
                         const Cl_options & cl_opts,
@@ -552,27 +556,12 @@ bool compare_file_contents( std::string & estr, std::string & ostr,
                             const char * const filename, const int infd2 );
 int decode( const Cl_options & cl_opts );
 
-// defined in decode_lz.cc
-struct Archive_descriptor;
-int decode_lz( const Cl_options & cl_opts, const Archive_descriptor & ad,
-               std::vector< char > & name_pending );
-
 // defined in delete.cc
-bool safe_seek( const int fd, const long long pos );
-int tail_copy( const Arg_parser & parser, const Archive_descriptor & ad,
-               std::vector< char > & name_pending, const long long istream_pos,
-               const int outfd, int retval );
 int delete_members( const Cl_options & cl_opts );
 
-// defined in delete_lz.cc
-int delete_members_lz( const Cl_options & cl_opts,
-                       const Archive_descriptor & ad,
-                       std::vector< char > & name_pending, const int outfd );
-
 // defined in exclude.cc
 namespace Exclude {
 void add_pattern( const std::string & arg );
-void clear();
 bool excluded( const char * const filename );
 } // end namespace Exclude
 
@@ -590,13 +579,15 @@ struct stat;
 int hstat( const char * const filename, struct stat * const st,
            const bool dereference );
 bool nonempty_arg( const Arg_parser & parser, const int i );
-int open_instream( const std::string & name );
+int open_instream( const char * const name, struct stat * const in_statsp = 0 );
 int open_outstream( const std::string & name, const bool create = true,
                     Resizable_buffer * const rbufp = 0, const bool force = true );
 void show_error( const char * const msg, const int errcode = 0,
                  const bool help = false );
 bool format_error( Resizable_buffer & rbuf, const int errcode,
                    const char * const format, ... );
+bool format_error( std::string & msg, const int errcode,
+                   const char * const format, ... );
 void print_error( const int errcode, const char * const format, ... );
 void format_file_error( std::string & estr, const char * const filename,
                         const char * const msg, const int errcode = 0 );
